@@ -13,23 +13,24 @@
  * and I'll send you a copy.
  *
  */
-
 #include <config.h>
 #include <includes.h>
 #include <freeradius-client.h>
+#include "util.h"
 
-/*
- * Function: rc_avpair_add
+/** Adds an attribute-value pair to the given list
  *
- * Purpose: add an attribute-value pair to the given list.
- *
- * Returns: pointer to added a/v pair upon success, NULL pointer upon failure.
- *
- * Remarks: Always appends the new pair to the end of the list.
+ * @note Always appends the new pair to the end of the list.
  *
+ * @param rh a handle to parsed configuration.
+ * @param list a #VALUE_PAIR array of values; initially must be %NULL.
+ * @param attrid The attribute of the pair to add (e.g., %PW_USER_NAME).
+ * @param pval the value (e.g., the actual username).
+ * @param len the length of @pval, or -1 if to calculate (in case of strings).
+ * @param vendorpec The vendor ID in case of a vendor specific value - 0 otherwise.
+ * @return pointer to added a/v pair upon success, NULL pointer upon failure.
  */
-
-VALUE_PAIR *rc_avpair_add (rc_handle const *rh, VALUE_PAIR **list, int attrid, void const *pval, int len, int vendorpec)
+VALUE_PAIR *rc_avpair_add (rc_handle const *rh, VALUE_PAIR **list, uint32_t attrid, void const *pval, int len, uint32_t vendorpec)
 {
 	VALUE_PAIR     *vp;
 
@@ -44,16 +45,13 @@ VALUE_PAIR *rc_avpair_add (rc_handle const *rh, VALUE_PAIR **list, int attrid, v
 
 }
 
-/*
- * Function: rc_avpair_assign
- *
- * Purpose: assign the given value to an attribute-value pair.
- *
- * Returns:  0 on success,
- *	    -1 on failure.
+/** Assigns the given value to an attribute-value pair
  *
+ * @param vp a pointer to a #VALUE_PAIR structure.
+ * @param pval the value (e.g., the actual username).
+ * @param len the length of @pval, or -1 if to calculate (in case of strings).
+ * @return 0 on success or -1 on failure.
  */
-
 int rc_avpair_assign (VALUE_PAIR *vp, void const *pval, int len)
 {
 
@@ -101,22 +99,21 @@ int rc_avpair_assign (VALUE_PAIR *vp, void const *pval, int len)
 	return 0;
 }
 
-/*
- * Function: rc_avpair_new
- *
- * Purpose: make a new attribute-value pair with given parameters.
- *
- * Returns: pointer to generated a/v pair when successful, NULL when failure.
+/** Make a new attribute-value pair with given parameters
  *
+ * @param rh a handle to parsed configuration.
+ * @param attrid The attribute of the pair to add (e.g., %PW_USER_NAME).
+ * @param pval the value (e.g., the actual username).
+ * @param len the length of pval, or -1 if to calculate (in case of strings).
+ * @param vendorpec The vendor ID in case of a vendor specific value - 0 otherwise.
+ * @return pointer to generated a/v pair when successful, NULL when failure.
  */
-
-VALUE_PAIR *rc_avpair_new (rc_handle const *rh, int attrid, void const *pval, int len, int vendorpec)
+VALUE_PAIR *rc_avpair_new (rc_handle const *rh, uint32_t attrid, void const *pval, int len, uint32_t vendorpec)
 {
 	VALUE_PAIR     *vp = NULL;
 	DICT_ATTR      *pda;
 
-	attrid = attrid | (vendorpec << 16);
-	if ((pda = rc_dict_getattr (rh, attrid)) == NULL)
+	if ((pda = rc_dict_get_vendor_attr(rh, attrid, vendorpec)) == NULL)
 	{
 		rc_log(LOG_ERR,"rc_avpair_new: unknown attribute %d", attrid);
 		return NULL;
@@ -128,7 +125,8 @@ VALUE_PAIR *rc_avpair_new (rc_handle const *rh, int attrid, void const *pval, in
 	}
 	if ((vp = malloc (sizeof (VALUE_PAIR))) != NULL)
 	{
-		strncpy (vp->name, pda->name, sizeof (vp->name));
+		strlcpy (vp->name, pda->name, sizeof (vp->name));
+		vp->vendor = vendorpec;
 		vp->attribute = attrid;
 		vp->next = NULL;
 		vp->type = pda->type;
@@ -171,23 +169,23 @@ VALUE_PAIR *rc_avpair_new (rc_handle const *rh, int attrid, void const *pval, in
 	return vp;
 }
 
-/*
+/** Takes attribute/value pairs from buffer and builds a value_pair list using allocated memory
  *
- * Function: rc_avpair_gen
+ * @note Uses recursion.
  *
- * Purpose: takes attribute/value pairs from buffer and builds a
- *	    value_pair list using allocated memory. Uses recursion.
- *
- * Returns: value_pair list or NULL on failure
+ * @param rh a handle to parsed configuration.
+ * @param pair a pointer to a #VALUE_PAIR structure.
+ * @param ptr the value (e.g., the actual username).
+ * @param length the length of ptr, or -1 if to calculate (in case of strings).
+ * @param vendorpec The vendor ID in case of a vendor specific value - 0 otherwise.
+ * @return value_pair list or %NULL on failure.
  */
-
-VALUE_PAIR *
-rc_avpair_gen(rc_handle const *rh, VALUE_PAIR *pair, unsigned char const *ptr,
-    int length, int vendorpec)
+VALUE_PAIR *rc_avpair_gen(rc_handle const *rh, VALUE_PAIR *pair, unsigned char const *ptr,
+			  int length, uint32_t vendorpec)
 {
-	int attribute, attrlen, x_len;
+	int attrlen, x_len;
 	unsigned char const *x_ptr;
-	uint32_t lvalue;
+	uint32_t attribute, lvalue;
 	DICT_ATTR *attr;
 	VALUE_PAIR *rpair;
 	char buffer[(AUTH_STRING_LEN * 2) + 1];
@@ -210,21 +208,21 @@ rc_avpair_gen(rc_handle const *rh, VALUE_PAIR *pair, unsigned char const *ptr,
 	if (length != attrlen) {
 		pair = rc_avpair_gen(rh, pair, ptr + attrlen, length - attrlen,
 		    vendorpec);
-		if (pair == NULL)
+		if ((pair == NULL) && (vendorpec != 0))
 			return NULL;
 	}
 
 	/* Actual processing */
-	attribute = ptr[0] | (vendorpec << 16);
+	attribute = ptr[0];
 	ptr += 2;
 	attrlen -= 2;
 
 	/* VSA */
-	if (attribute == PW_VENDOR_SPECIFIC) {
+	if ((vendorpec == 0) && (attribute == PW_VENDOR_SPECIFIC)) {
 		if (attrlen < 4) {
 			rc_log(LOG_ERR, "rc_avpair_gen: received VSA "
 			    "attribute with invalid length");
-			goto shithappens;
+			goto skipit;
 		}
 		memcpy(&lvalue, ptr, 4);
 		vendorpec = ntohl(lvalue);
@@ -232,7 +230,7 @@ rc_avpair_gen(rc_handle const *rh, VALUE_PAIR *pair, unsigned char const *ptr,
 			/* Warn and skip over the unknown VSA */
 			rc_log(LOG_WARNING, "rc_avpair_gen: received VSA "
 			    "attribute with unknown Vendor-Id %d", vendorpec);
-			return pair;
+			goto skipit;
 		}
 		/* Process recursively */
 		return rc_avpair_gen(rh, pair, ptr + 4, attrlen - 4,
@@ -240,7 +238,7 @@ rc_avpair_gen(rc_handle const *rh, VALUE_PAIR *pair, unsigned char const *ptr,
 	}
 
 	/* Normal */
-	attr = rc_dict_getattr(rh, attribute);
+	attr = rc_dict_get_vendor_attr(rh, attribute, vendorpec);
 	if (attr == NULL) {
 		buffer[0] = '\0';	/* Initial length. */
 		x_ptr = ptr;
@@ -255,10 +253,10 @@ rc_avpair_gen(rc_handle const *rh, VALUE_PAIR *pair, unsigned char const *ptr,
 		} else {
 			rc_log(LOG_WARNING, "rc_avpair_gen: received "
 			    "unknown VSA attribute %d, vendor %d of "
-			    "length %d: 0x%s", attribute & 0xffff,
-			    VENDOR(attribute), attrlen + 2, buffer);
+			    "length %d: 0x%s", attribute,
+			    vendorpec, attrlen + 2, buffer);
 		}
-		goto shithappens;
+		goto skipit;
 	}
 
 	rpair = malloc(sizeof(*rpair));
@@ -272,6 +270,7 @@ rc_avpair_gen(rc_handle const *rh, VALUE_PAIR *pair, unsigned char const *ptr,
 	rpair->next = pair;
 	pair = rpair;
 	strcpy(pair->name, attr->name);
+	pair->vendor = attr->vendor;
 	pair->attribute = attr->value;
 	pair->type = attr->type;
 
@@ -286,13 +285,13 @@ rc_avpair_gen(rc_handle const *rh, VALUE_PAIR *pair, unsigned char const *ptr,
 		if (attrlen != 4) {
 			rc_log(LOG_ERR, "rc_avpair_gen: received INT "
 			    "attribute with invalid length");
-			goto shithappens;
+			goto skipit;
 		}
 	case PW_TYPE_IPADDR:
 		if (attrlen != 4) {
 			rc_log(LOG_ERR, "rc_avpair_gen: received IPADDR"
 			    " attribute with invalid length");
-			goto shithappens;
+			goto skipit;
 		}
 		memcpy((char *)&lvalue, (char *)ptr, 4);
 		pair->lvalue = ntohl(lvalue);
@@ -301,7 +300,7 @@ rc_avpair_gen(rc_handle const *rh, VALUE_PAIR *pair, unsigned char const *ptr,
 		if (attrlen != 16) {
 			rc_log(LOG_ERR, "rc_avpair_gen: received IPV6ADDR"
 			    " attribute with invalid length");
-			goto shithappens;
+			goto skipit;
 		}
 		memcpy(pair->strvalue, (char *)ptr, 16);
 		pair->lvalue = attrlen;
@@ -310,7 +309,7 @@ rc_avpair_gen(rc_handle const *rh, VALUE_PAIR *pair, unsigned char const *ptr,
 		if (attrlen > 18 || attrlen < 2) {
 			rc_log(LOG_ERR, "rc_avpair_gen: received IPV6PREFIX"
 			    " attribute with invalid length: %d", attrlen);
-			goto shithappens;
+			goto skipit;
 		}
 		memcpy(pair->strvalue, (char *)ptr, attrlen);
 		pair->lvalue = attrlen;
@@ -319,14 +318,16 @@ rc_avpair_gen(rc_handle const *rh, VALUE_PAIR *pair, unsigned char const *ptr,
 		if (attrlen != 4) {
 			rc_log(LOG_ERR, "rc_avpair_gen: received DATE "
 			    "attribute with invalid length");
-			goto shithappens;
+			goto skipit;
 		}
 
 	default:
 		rc_log(LOG_WARNING, "rc_avpair_gen: %s has unknown type",
 		    attr->name);
-		goto shithappens;
+		goto skipit;
 	}
+
+skipit:
 	return pair;
 
 shithappens:
@@ -338,37 +339,33 @@ shithappens:
 	return NULL;
 }
 
-/*
- * Function: rc_avpair_get
- *
- * Purpose: Find the first attribute value-pair (which matches the given
- *          attribute) from the specified value-pair list.
- *
- * Returns: found value_pair
+/** Find the first attribute value-pair (which matches the given attribute) from the specified value-pair list
  *
+ * @param vp a pointer to a #VALUE_PAIR structure.
+ * @param attrid The attribute of the pair to find (e.g., %PW_USER_NAME).
+ * @param vendorpec The vendor ID in case of a vendor specific value - 0 otherwise.
+ * @return the value pair found.
  */
-
-VALUE_PAIR *rc_avpair_get (VALUE_PAIR *vp, int attrid, int vendorpec)
+VALUE_PAIR *rc_avpair_get (VALUE_PAIR *vp, uint32_t attrid, uint32_t vendorpec)
 {
-	for (; vp != NULL && !(ATTRID(vp->attribute) == ATTRID(attrid) &&
-	    VENDOR(vp->attribute) == vendorpec); vp = vp->next)
+	for (; vp != NULL && !(vp->attribute == attrid &&
+	    vp->vendor == vendorpec); vp = vp->next)
 	{
 		continue;
 	}
 	return vp;
 }
 
-/*
- * Function: rc_avpair_insert
+/** Insert a VALUE_PAIR into a list
  *
- * Purpose: Given the address of an existing list "a" and a pointer
- *	    to an entry "p" in that list, add the value pair "b" to
- *	    the "a" list after the "p" entry.  If "p" is NULL, add
- *	    the value pair "b" to the end of "a".
+ * Given the address of an existing list "a" and a pointer to an entry "p" in that list, add the value pair "b" to
+ * the "a" list after the "p" entry.  If "p" is NULL, add the value pair "b" to the end of "a".
  *
+ * @param a a #VALUE_PAIR array of values.
+ * @param p a pointer to a #VALUE_PAIR in a.
+ * @param b The #VALUE_PAIR pointer to add in a.
  */
-
-void rc_avpair_insert (VALUE_PAIR **a, VALUE_PAIR *p, VALUE_PAIR *b)
+void rc_avpair_insert(VALUE_PAIR **a, VALUE_PAIR *p, VALUE_PAIR *b)
 {
 	VALUE_PAIR     *this_node = NULL;
 	VALUE_PAIR     *vp;
@@ -414,13 +411,10 @@ void rc_avpair_insert (VALUE_PAIR **a, VALUE_PAIR *p, VALUE_PAIR *b)
 	return;
 }
 
-/*
- * Function: rc_avpair_free
- *
- * Purpose: frees all value_pairs in the list
+/** Frees all value_pairs in the list
  *
+ * @param pair a pointer to a VALUE_PAIR structure.
  */
-
 void rc_avpair_free (VALUE_PAIR *pair)
 {
 	VALUE_PAIR     *next;
@@ -433,18 +427,17 @@ void rc_avpair_free (VALUE_PAIR *pair)
 	}
 }
 
-/*
- * Function: rc_fieldcpy
+/** Copy a data field from the buffer
  *
- * Purpose: Copy a data field from the buffer.  Advance the buffer
- *          past the data field. Ensure that no more than len - 1
- *          bytes are copied and that resulting string is terminated
- *          with '\0'.
+ * Advance the buffer past the data field. Ensure that no more than len - 1 bytes are copied and that resulting
+ * string is terminated with '\0'.
  *
+ * @param string the provided string to copy.
+ * @param uptr the current position of the buffer.
+ * @param stopat characters to which parsing should stop.
+ * @param len the maximum length of string.
  */
-
-static void
-rc_fieldcpy(char *string, char const **uptr, char const *stopat, size_t len)
+static void rc_fieldcpy(char *string, char const **uptr, char const *stopat, size_t len)
 {
 	char const *ptr, *estring;
 
@@ -479,22 +472,18 @@ rc_fieldcpy(char *string, char const **uptr, char const *stopat, size_t len)
 	return;
 }
 
-
-/*
- * Function: rc_avpair_parse
- *
- * Purpose: parses the buffer to extract the attribute-value pairs.
- *
- * Returns: 0 = successful parse of attribute-value pair,
- *	   -1 = syntax (or other) error detected.
- *
- */
-
 #define PARSE_MODE_NAME		0
 #define PARSE_MODE_EQUAL	1
 #define PARSE_MODE_VALUE	2
 #define PARSE_MODE_INVALID	3
 
+/** Parses the buffer to extract the attribute-value pairs
+ *
+ * @param rh a handle to parsed configuration.
+ * @param buffer the buffer to be parsed.
+ * @param first_pair an allocated array of values.
+ * @return 0 on successful parse of attribute-value pair, or -1 on syntax (or other) error detected.
+ */
 int rc_avpair_parse (rc_handle const *rh, char const *buffer, VALUE_PAIR **first_pair)
 {
 	int             mode;
@@ -565,6 +554,7 @@ int rc_avpair_parse (rc_handle const *rh, char const *buffer, VALUE_PAIR **first
 			strcpy (pair->name, attr->name);
 			pair->attribute = attr->value;
 			pair->type = attr->type;
+			pair->vendor = attr->vendor;
 
 			switch (pair->type)
 			{
@@ -600,7 +590,13 @@ int rc_avpair_parse (rc_handle const *rh, char const *buffer, VALUE_PAIR **first
 				break;
 
 			    case PW_TYPE_IPADDR:
-                                pair->lvalue = rc_get_ipaddr(valstr);
+			    	if (inet_pton(AF_INET, valstr, &pair->lvalue) == 0) {
+			    		rc_log(LOG_ERR, "rc_avpair_parse: invalid IPv4 address %s", valstr);
+			    		free(pair);
+			    		return -1;
+			    	}
+
+                                pair->lvalue = ntohl(pair->lvalue);
 				break;
 
 			    case PW_TYPE_IPV6ADDR:
@@ -706,21 +702,23 @@ int rc_avpair_parse (rc_handle const *rh, char const *buffer, VALUE_PAIR **first
 	return 0;
 }
 
-/*
- * Function: rc_avpair_tostr
- *
- * Purpose: Translate an av_pair into two strings
- *
- * Returns: 0 on success, -1 on failure
+/** Translate an av_pair into two strings
  *
+ * @param rh a handle to parsed configuration.
+ * @param pair a pointer to a VALUE_PAIR structure.
+ * @param name the name of the pair.
+ * @param ln the size of name.
+ * @param value the value of the pair.
+ * @param lv the size of value.
+ * @return 0 on success, -1 on failure.
  */
-
 int rc_avpair_tostr (rc_handle const *rh, VALUE_PAIR *pair, char *name, int ln, char *value, int lv)
 {
 	DICT_VALUE     *dval;
 	char            buffer[32];
 	struct in_addr  inad;
-	unsigned char         *ptr;
+	unsigned char  *ptr;
+	unsigned int    pos;
 
 	*name = *value = '\0';
 
@@ -729,12 +727,13 @@ int rc_avpair_tostr (rc_handle const *rh, VALUE_PAIR *pair, char *name, int ln,
 		return -1;
 	}
 
-	strncpy(name, pair->name, (size_t) ln);
+	strlcpy(name, pair->name, (size_t) ln);
 
 	switch (pair->type)
 	{
 	    case PW_TYPE_STRING:
 	    	lv--;
+	    	pos = 0;
 		ptr = (unsigned char *) pair->strvalue;
 		if (pair->attribute == PW_DIGEST_ATTRIBUTES) {
 			pair->strvalue[*(ptr + 1)] = '\0';
@@ -744,47 +743,53 @@ int rc_avpair_tostr (rc_handle const *rh, VALUE_PAIR *pair, char *name, int ln,
 		{
 			if (!(isprint (*ptr)))
 			{
-				snprintf (buffer, sizeof(buffer), "\\%03o", *ptr);
-				strncat(value, buffer, (size_t) lv);
-				lv -= 4;
-				if (lv < 0) break;
+				if (lv >= 4) {
+					snprintf (&value[pos], lv, "\\%03o", *ptr);
+					pos += 4;
+					lv -= 4;
+				} else {
+					break;
+				}
 			}
 			else
 			{
-				strncat(value, (char *)ptr, 1);
-				lv--;
-				if (lv <= 0) break;
+				if (lv > 0) {
+					value[pos++] = *ptr;
+					lv--;
+				} else {
+					break;
+				}
 			}
 			ptr++;
 		}
+		value[pos] = 0;
 		break;
 
 	    case PW_TYPE_INTEGER:
 		dval = rc_dict_getval (rh, pair->lvalue, pair->name);
 		if (dval != NULL)
 		{
-			strncpy(value, dval->name, (size_t) lv-1);
+			strlcpy(value, dval->name, (size_t) lv);
 		}
 		else
 		{
-			snprintf(buffer, sizeof(buffer), "%ld", (long int)pair->lvalue);
-			strncpy(value, buffer, (size_t) lv);
+			snprintf(value, lv, "%ld", (long int)pair->lvalue);
 		}
 		break;
 
 	    case PW_TYPE_IPADDR:
 		inad.s_addr = htonl(pair->lvalue);
-		strncpy (value, inet_ntoa (inad), (size_t) lv-1);
+		strlcpy (value, inet_ntoa (inad), (size_t) lv);
 		break;
 
 	    case PW_TYPE_IPV6ADDR:
-	    	if (inet_ntop(AF_INET6, pair->strvalue, value, lv-1) == NULL)
+	    	if (inet_ntop(AF_INET6, pair->strvalue, value, lv) == NULL)
 	    		return -1;
 		break;
 
 	    case PW_TYPE_IPV6PREFIX: {
 	    	uint8_t ip[16];
-	    	uint8_t txt[48];
+	    	char txt[48];
 	    	if (pair->lvalue < 2)
 	    		return -1;
 
@@ -793,15 +798,24 @@ int rc_avpair_tostr (rc_handle const *rh, VALUE_PAIR *pair, char *name, int ln,
 
 	    	if (inet_ntop(AF_INET6, ip, txt, sizeof(txt)) == NULL)
 	    		return -1;
-		snprintf(value, lv-1, "%s/%u", txt, (unsigned)pair->strvalue[1]);
+		snprintf(value, lv, "%s/%u", txt, (unsigned)pair->strvalue[1]);
 
 		break;
 	    }
-	    case PW_TYPE_DATE:
-		strftime (buffer, sizeof (buffer), "%m/%d/%y %H:%M:%S",
-			  gmtime ((time_t *) & pair->lvalue));
-		strncpy(value, buffer, lv-1);
+	    case PW_TYPE_DATE: {
+		struct tm * ptm = gmtime((time_t *) & pair->lvalue);
+		if (!ptm) /* NOTE: It is possible for gmtime to return NULL, so check for it */
+		{
+			/* The time value is bad, so we will use the current time instead. The alternative would be to return "01/01/70 00:00:00" */
+			rc_log(LOG_WARNING, "rc_avpair_tostr: bad time value, using current time instead. errno = %d", errno);
+			time_t now;
+			time(&now);
+			ptm = gmtime(&now);
+		}
+
+		strftime(value, lv, "%m/%d/%y %H:%M:%S", ptm);
 		break;
+	    }
 
 	    default:
 		rc_log(LOG_ERR, "rc_avpair_tostr: unknown attribute type %d", pair->type);
@@ -812,16 +826,17 @@ int rc_avpair_tostr (rc_handle const *rh, VALUE_PAIR *pair, char *name, int ln,
 	return 0;
 }
 
-/*
- * Function: rc_avpair_log
+/** Format a sequence of attribute value pairs into a printable string
  *
- * Purpose: format sequence of attribute value pairs into printable
- * string. The caller should provide a storage buffer and the buffer length.
- * Returns pointer to provided storage buffer.
+ * The caller should provide a storage buffer and the buffer length.
  *
+ * @param rh a handle to parsed configuration.
+ * @param pair a pointer to a #VALUE_PAIR structure.
+ * @param buf will hold the string output of the pair.
+ * @param len the size of buf.
+ * @return a pointer to provided storage buffer.
  */
-char *
-rc_avpair_log(rc_handle const *rh, VALUE_PAIR *pair, char *buf, size_t buf_len)
+char *rc_avpair_log(rc_handle const *rh, VALUE_PAIR *pair, char *buf, size_t buf_len)
 {
 	size_t len, nlen;
 	VALUE_PAIR *vp;
@@ -841,14 +856,12 @@ rc_avpair_log(rc_handle const *rh, VALUE_PAIR *pair, char *buf, size_t buf_len)
 	return buf;
 }
 
-/*
- * Function: rc_avpair_readin
- *
- * Purpose: get a sequence of attribute value pairs from the file input
- *	    and make them into a list of value_pairs
+/** Get a sequence of attribute value pairs from the file input and make them into a list of value_pairs
  *
+ * @param rh a handle to parsed configuration.
+ * @param input a %FILE handle.
+ * @return the array of value pairs.
  */
-
 VALUE_PAIR *rc_avpair_readin(rc_handle const *rh, FILE *input)
 {
 	VALUE_PAIR *vp = NULL;
