@@ -18,15 +18,17 @@
 #include <includes.h>
 #include <freeradius-client.h>
 #include <options.h>
+#include "util.h"
 
-/*
- * Function: find_option
+#define SA(p)	((struct sockaddr *)(p))
+
+/** Find an option in the option list
  *
- * Purpose: find an option in the option list
- *
- * Returns: pointer to option on success, NULL otherwise
+ * @param rh a handle to parsed configuration.
+ * @param optname the name of the option.
+ * @param type the option type.
+ * @return pointer to option on success, NULL otherwise.
  */
-
 static OPTION *find_option(rc_handle const *rh, char const *optname, unsigned int type)
 {
 	int 	i;
@@ -43,14 +45,13 @@ static OPTION *find_option(rc_handle const *rh, char const *optname, unsigned in
 	return NULL;
 }
 
-/*
- * Function: set_option_...
+/** Set a specific option doing type conversions
  *
- * Purpose: set a specific option doing type conversions
- *
- * Returns: 0 on success, -1 on failure
+ * @param filename the name of the config file (for logging purposes).
+ * @param line the line number in the file.
+ * @param p Value.
+ * @return 0 on success, -1 on failure.
  */
-
 static int set_option_str(char const *filename, int line, OPTION *option, char const *p)
 {
 	if (p) {
@@ -118,7 +119,41 @@ static int set_option_srv(char const *filename, int line, OPTION *option, char c
 
 	p_pointer = strtok_r(p_dupe, ", \t", &p_save);
 
-	/* Check to see if we have 'servername:port' syntax */
+	/* check to see for '[IPv6]:port' syntax */
+	if ((q = strchr(p_pointer,'[')) != NULL) {
+		*q = '\0';
+		q++;
+		p_pointer = q;
+
+		q = strchr(p_pointer, ']');
+		if (q == NULL) {
+			free(p_dupe);
+			rc_log(LOG_CRIT, "read_config: IPv6 parse error");
+			return -1;
+		}
+		*q = '\0';
+		q++;
+
+		if (q[0] == ':') {
+			q++;
+		}
+
+		/* Check to see if we have '[IPv6]:port:secret' syntax */
+		if((s=strchr(q, ':')) != NULL) {
+			*s = '\0';
+			s++;
+			serv->secret[serv->max] = strdup(s);
+			if (serv->secret[serv->max] == NULL) {
+				rc_log(LOG_CRIT, "read_config: out of memory");
+				if (option->val == NULL) {
+					free(p_dupe);
+					free(serv);
+				}
+				return -1;
+			}
+		}
+
+	} else /* Check to see if we have 'servername:port' syntax */
 	if ((q = strchr(p_pointer,':')) != NULL) {
 		*q = '\0';
 		q++;
@@ -138,6 +173,7 @@ static int set_option_srv(char const *filename, int line, OPTION *option, char c
 			}
 		}
 	}
+
 	if(q && strlen(q) > 0) {
 		serv->port[serv->max] = atoi(q);
 	} else {
@@ -202,9 +238,7 @@ static int set_option_auo(char const *filename, int line, OPTION *option, char c
 	}
 
 	*iptr = 0;
-	/*if(strstr(p_dupe,", \t") != NULL) {*/
-		p_pointer = strtok_r(p_dupe, ", \t", &p_save);
-	/*}*/
+	p_pointer = strtok_r(p_dupe, ", \t", &p_save);
 
 	if (!strncmp(p_pointer, "local", 5))
 			*iptr = AUTH_LOCAL_FST;
@@ -238,14 +272,15 @@ static int set_option_auo(char const *filename, int line, OPTION *option, char c
 	return 0;
 }
 
-
-/* Function: rc_add_config
+/** Allow a config option to be added to rc_handle from inside a program
  *
- * Purpose: allow a config option to be added to rc_handle from inside a program
- *
- * Returns: 0 on success, -1 on failure
+ * @param rh a handle to parsed configuration.
+ * @param option_name the name of the option.
+ * @param option_val the value to be added.
+ * @param source typically should be %__FILE__ or %__func__ for logging purposes.
+ * @param line %__LINE__ for logging purposes.
+ * @return 0 on success, -1 on failure.
  */
-
 int rc_add_config(rc_handle *rh, char const *option_name, char const *option_val, char const *source, int line)
 {
 	OPTION *option;
@@ -287,20 +322,26 @@ int rc_add_config(rc_handle *rh, char const *option_name, char const *option_val
 			rc_log(LOG_CRIT, "rc_add_config: impossible case branch!");
 			abort();
 	}
+
+	if (strcmp(option->name, "bindaddr") == 0) {
+		memset(&rh->own_bind_addr, 0, sizeof(rh->own_bind_addr));
+		rh->own_bind_addr_set = 0;
+		rc_own_bind_addr(rh, &rh->own_bind_addr);
+		rh->own_bind_addr_set = 1;
+	}
+
 	return 0;
 }
 
-/*
- * Function: rc_config_init
+/** Initialise a configuration structure
  *
- * Purpose: initialize the configuration structure from an external program.  For use when not
+ * Initialize the configuration structure from an external program.  For use when not
  * running a standalone client that reads from a config file.
  *
- * Returns: rc_handle on success, NULL on failure
+ * @param rh a handle to parsed configuration.
+ * @return rc_handle on success, NULL on failure.
  */
-
-rc_handle *
-rc_config_init(rc_handle *rh)
+rc_handle *rc_config_init(rc_handle *rh)
 {
 	int i;
 	SERVER *authservers;
@@ -347,17 +388,12 @@ rc_config_init(rc_handle *rh)
 	return rh;
 }
 
-
-/*
- * Function: rc_read_config
+/** Read the global config file
  *
- * Purpose: read the global config file
- *
- * Returns: new rc_handle on success, NULL when failure
+ * @param filename a name of a file.
+ * @return new rc_handle on success, NULL when failure.
  */
-
-rc_handle *
-rc_read_config(char const *filename)
+rc_handle *rc_read_config(char const *filename)
 {
 	FILE *configfd;
 	char buffer[512], *p;
@@ -366,8 +402,6 @@ rc_read_config(char const *filename)
 	size_t pos;
 	rc_handle *rh;
 
-	srandom((unsigned int)(time(NULL)+getpid()));
-
 	rh = rc_new();
 	if (rh == NULL)
 		return NULL;
@@ -473,14 +507,12 @@ rc_read_config(char const *filename)
 	return rh;
 }
 
-/*
- * Function: rc_conf_str, rc_conf_int, rc_conf_src
+/** Get the value of a config option
  *
- * Purpose: get the value of a config option
- *
- * Returns: config option value
+ * @param rh a handle to parsed configuration.
+ * @param optname the name of an option.
+ * @return config option value.
  */
-
 char *rc_conf_str(rc_handle const *rh, char const *optname)
 {
 	OPTION *option;
@@ -490,12 +522,17 @@ char *rc_conf_str(rc_handle const *rh, char const *optname)
 	if (option != NULL) {
 		return (char *)option->val;
 	} else {
-		rc_log(LOG_CRIT, "rc_conf_str: unkown config option requested: %s", optname);
-		abort();
+		rc_log(LOG_CRIT, "rc_conf_str: unknown config option requested: %s", optname);
 		return NULL;
 	}
 }
 
+/** Get the value of a config option
+ *
+ * @param rh a handle to parsed configuration.
+ * @param optname the name of an option.
+ * @return config option value.
+ */
 int rc_conf_int(rc_handle const *rh, char const *optname)
 {
 	OPTION *option;
@@ -510,12 +547,17 @@ int rc_conf_int(rc_handle const *rh, char const *optname)
 			return 0;
 		}
 	} else {
-		rc_log(LOG_CRIT, "rc_conf_int: unkown config option requested: %s", optname);
-		abort();
+		rc_log(LOG_CRIT, "rc_conf_int: unknown config option requested: %s", optname);
 		return 0;
 	}
 }
 
+/** Get the value of a config option
+ *
+ * @param rh a handle to parsed configuration.
+ * @param optname the name of an option.
+ * @return config option value.
+ */
 SERVER *rc_conf_srv(rc_handle const *rh, char const *optname)
 {
 	OPTION *option;
@@ -525,33 +567,30 @@ SERVER *rc_conf_srv(rc_handle const *rh, char const *optname)
 	if (option != NULL) {
 		return (SERVER *)option->val;
 	} else {
-		rc_log(LOG_CRIT, "rc_conf_srv: unkown config option requested: %s", optname);
-		abort();
+		rc_log(LOG_CRIT, "rc_conf_srv: unknown config option requested: %s", optname);
 		return NULL;
 	}
 }
 
-/*
- * Function: test_config
+/** Tests the configuration the user supplied
  *
- * Purpose: test the configuration the user supplied
- *
- * Returns: 0 on success, -1 when failure
+ * @param rh a handle to parsed configuration.
+ * @param filename a name of a configuration file.
+ * @return 0 on success, -1 when failure.
  */
-
 int test_config(rc_handle const *rh, char const *filename)
 {
-#if 0
-	struct stat st;
-	char	    *file;
-#endif
+	SERVER *srv;
 
-	if (!(rc_conf_srv(rh, "authserver")->max))
+	srv = rc_conf_srv(rh, "authserver");
+	if (!srv || !srv->max)
 	{
 		rc_log(LOG_ERR,"%s: no authserver specified", filename);
 		return -1;
 	}
-	if (!(rc_conf_srv(rh, "acctserver")->max))
+
+	srv = rc_conf_srv(rh, "acctserver");
+	if (!srv || !srv->max)
 	{
 		rc_log(LOG_ERR,"%s: no acctserver specified", filename);
 		return -1;
@@ -582,41 +621,11 @@ int test_config(rc_handle const *rh, char const *filename)
 		rc_log(LOG_ERR,"%s: radius_deadtime is illegal", filename);
 		return -1;
 	}
-#if 0
-	file = rc_conf_str(rh, "login_local");
-	if (stat(file, &st) == 0)
-	{
-		if (!S_ISREG(st.st_mode)) {
-			rc_log(LOG_ERR,"%s: not a regular file: %s", filename, file);
-			return -1;
-		}
-	} else {
-		rc_log(LOG_ERR,"%s: file not found: %s", filename, file);
-		return -1;
-	}
-	file = rc_conf_str(rh, "login_radius");
-	if (stat(file, &st) == 0)
-	{
-		if (!S_ISREG(st.st_mode)) {
-			rc_log(LOG_ERR,"%s: not a regular file: %s", filename, file);
-			return -1;
-		}
-	} else {
-		rc_log(LOG_ERR,"%s: file not found: %s", filename, file);
-		return -1;
-	}
-#endif
-
 	if (rc_conf_int(rh, "login_tries") <= 0)
 	{
 		rc_log(LOG_ERR,"%s: login_tries <= 0 is illegal", filename);
 		return -1;
 	}
-	if (rc_conf_str(rh, "seqfile") == NULL)
-	{
-		rc_log(LOG_ERR,"%s: seqfile not specified", filename);
-		return -1;
-	}
 	if (rc_conf_int(rh, "login_timeout") <= 0)
 	{
 		rc_log(LOG_ERR,"%s: login_timeout <= 0 is illegal", filename);
@@ -636,70 +645,58 @@ int test_config(rc_handle const *rh, char const *filename)
 	return 0;
 }
 
-/*
- * Function: rc_find_match
- *
- * Purpose: see if ip_addr is one of the ip addresses of hostname
- *
- * Returns: 0 on success, -1 when failure
+/** See if info matches hostname
  *
+ * @param info a struct addrinfo
+ * @param hostname the name of the host.
+ * @return 0 on success, -1 when failure.
  */
-
-static int find_match (uint32_t *ip_addr, char const *hostname)
+static int find_match (const struct addrinfo* addr, const struct addrinfo *hostname)
 {
+	const struct addrinfo *ptr, *ptr2;
+	unsigned len1, len2;
 
-	uint32_t           addr;
-	char          **paddr;
-	struct hostent *hp;
+	ptr = addr;
+	while(ptr) {
+		ptr2 = hostname;
+		while(ptr2) {
+			len1 = SA_GET_INLEN(ptr->ai_addr);
+			len2 = SA_GET_INLEN(ptr2->ai_addr);
 
-	if (rc_good_ipaddr (hostname) == 0)
-	{
-		if (*ip_addr == ntohl(inet_addr (hostname)))
-		{
-			return 0;
-		}
-		return -1;
-	}
-
-	if ((hp = rc_gethostbyname(hostname)) == NULL)
-	{
-		return -1;
-	}
-
-	for (paddr = hp->h_addr_list; *paddr; paddr++)
-	{
-		addr = ** (uint32_t **) paddr;
-		if (ntohl(addr) == *ip_addr)
-		{
-			return 0;
-		}
-	}
-	return -1;
+			if (len1 > 0 && 
+			    len1 == len2 && 
+			    memcmp(SA_GET_INADDR(ptr->ai_addr), SA_GET_INADDR(ptr2->ai_addr), len1) == 0) {
+				return 0;
+			}
+			ptr2 = ptr2->ai_next;
+ 		}
+		ptr = ptr->ai_next;
+ 	}
+ 	return -1;
 }
 
-/*
- * Function: rc_ipaddr_local
- *
- * Purpose: checks if provided address is local address
- *
- * Returns: 0 if local, 1 if not local, -1 on failure
+/** Checks if provided address is local address
  *
+ * @param addr an %AF_INET or %AF_INET6 address
+ * @return 0 if local, 1 if not local, -1 on failure.
  */
-
-static int
-rc_ipaddr_local(uint32_t ip_addr)
+static int rc_ipaddr_local(const struct sockaddr *addr)
 {
 	int temp_sock, res, serrno;
-	struct sockaddr_in sin;
+	struct sockaddr_storage tmpaddr;
 
-	temp_sock = socket(AF_INET, SOCK_DGRAM, 0);
+	memcpy(&tmpaddr, addr, SA_LEN(addr));
+
+	temp_sock = socket(addr->sa_family, SOCK_DGRAM, 0);
 	if (temp_sock == -1)
 		return -1;
-	memset(&sin, '\0', sizeof(sin));
-	sin.sin_family = AF_INET;
-	sin.sin_addr.s_addr = htonl(ip_addr);
-	sin.sin_port = htons(0);
-	res = bind(temp_sock, (struct sockaddr *)&sin, sizeof(sin));
+
+	if (addr->sa_family == AF_INET) {
+		((struct sockaddr_in*)&tmpaddr)->sin_port = 0;
+	} else {
+		((struct sockaddr_in6*)&tmpaddr)->sin6_port = 0;
+	}
+	res = bind(temp_sock, SA(&tmpaddr), SS_LEN(&tmpaddr));
 	serrno = errno;
 	close(temp_sock);
 	if (res == 0)
@@ -709,50 +706,41 @@ rc_ipaddr_local(uint32_t ip_addr)
 	return -1;
 }
 
-/*
- * Function: rc_is_myname
- *
- * Purpose: check if provided name refers to ourselves
- *
- * Returns: 0 if yes, 1 if no and -1 on failure
+/** Checks if provided name refers to ourselves
  *
+ * @param info an addrinfo of the host to check
+ * @return 0 if yes, 1 if no and -1 on failure.
  */
-
-static int
-rc_is_myname(char const *hostname)
+static int rc_is_myname(const struct addrinfo *info)
 {
-	uint32_t 	addr;
-	char 	**paddr;
-	struct 	hostent *hp;
+	const struct addrinfo *p;
 	int	res;
 
-	if (rc_good_ipaddr(hostname) == 0)
-		return rc_ipaddr_local(ntohl(inet_addr(hostname)));
-
-	if ((hp = rc_gethostbyname(hostname)) == NULL)
-		return -1;
-	for (paddr = hp->h_addr_list; *paddr; paddr++) {
-		addr = **(uint32_t **)paddr;
-		res = rc_ipaddr_local(ntohl(addr));
-		if (res == 0 || res == -1)
-			return res;
-	}
-	return 1;
+	p = info;
+	while(p != NULL) {
+		res = rc_ipaddr_local(p->ai_addr);
+		if (res == 0 || res == -1) {
+ 			return res;
+		}
+		p = p->ai_next;
+ 	}
+ 	return 1;
 }
 
-/*
- * Function: rc_find_server
- *
- * Purpose: locate a server in the rh config or if not found, check for a servers file
- *
- * Returns: 0 on success, -1 on failure
+/** Locate a server in the rh config or if not found, check for a servers file
  *
+ * @param rh a handle to parsed configuration.
+ * @param server_name the name of the server.
+ * @param info: will hold a pointer to addrinfo
+ * @param secret will hold the server's secret (of %MAX_SECRET_LENGTH).
+ * @param flags %AUTH or %ACCT
+ 
+ * @return 0 on success, -1 on failure.
  */
-
-int rc_find_server (rc_handle const *rh, char const *server_name, uint32_t *ip_addr, char *secret)
+int rc_find_server_addr (rc_handle const *rh, char const *server_name,
+                         struct addrinfo** info, char *secret, unsigned flags)
 {
 	int		i;
-	size_t          len;
 	int             result = 0;
 	FILE           *clientfd;
 	char           *h;
@@ -763,48 +751,39 @@ int rc_find_server (rc_handle const *rh, char const *server_name, uint32_t *ip_a
 	char	       *hostnm_save;
 	SERVER	       *authservers;
 	SERVER	       *acctservers;
+	struct addrinfo *tmpinfo = NULL;
 
 	/* Lookup the IP address of the radius server */
-	if ((*ip_addr = rc_get_ipaddr (server_name)) == (uint32_t) 0)
+	if ((*info = rc_getaddrinfo (server_name, flags==AUTH?PW_AI_AUTH:PW_AI_ACCT)) == NULL)
 		return -1;
 
-	/* Check to see if the server secret is defined in the rh config */
-	if( (authservers = rc_conf_srv(rh, "authserver")) != NULL )
-	{
-		for( i = 0; i < authservers->max; i++ )
+	if (flags == AUTH) {
+		/* Check to see if the server secret is defined in the rh config */
+		if( (authservers = rc_conf_srv(rh, "authserver")) != NULL )
 		{
-			if( (strncmp(server_name, authservers->name[i], strlen(server_name)) == 0) &&
-			    (authservers->secret[i] != NULL) )
+			for( i = 0; i < authservers->max; i++ )
 			{
-				memset (secret, '\0', MAX_SECRET_LENGTH);
-				len = strlen (authservers->secret[i]);
-				if (len > MAX_SECRET_LENGTH)
+				if( (strncmp(server_name, authservers->name[i], strlen(server_name)) == 0) &&
+				    (authservers->secret[i] != NULL) )
 				{
-					len = MAX_SECRET_LENGTH;
+					memset (secret, '\0', MAX_SECRET_LENGTH);
+					strlcpy (secret, authservers->secret[i], MAX_SECRET_LENGTH);
+					return 0;
 				}
-				strncpy (secret, authservers->secret[i], (size_t) len);
-				secret[MAX_SECRET_LENGTH] = '\0';
-				return 0;
 			}
 		}
-	}
-
-	if( (acctservers = rc_conf_srv(rh, "acctserver")) != NULL )
-	{
-		for( i = 0; i < acctservers->max; i++ )
+	} else if (flags == ACCT) {
+		if( (acctservers = rc_conf_srv(rh, "acctserver")) != NULL )
 		{
-			if( (strncmp(server_name, acctservers->name[i], strlen(server_name)) == 0) &&
-			    (acctservers->secret[i] != NULL) )
+			for( i = 0; i < acctservers->max; i++ )
 			{
-				memset (secret, '\0', MAX_SECRET_LENGTH);
-				len = strlen (acctservers->secret[i]);
-				if (len > MAX_SECRET_LENGTH)
+				if( (strncmp(server_name, acctservers->name[i], strlen(server_name)) == 0) &&
+				    (acctservers->secret[i] != NULL) )
 				{
-					len = MAX_SECRET_LENGTH;
+					memset (secret, '\0', MAX_SECRET_LENGTH);
+					strlcpy (secret, acctservers->secret[i], MAX_SECRET_LENGTH);
+					return 0;
 				}
-				strncpy (secret, acctservers->secret[i], (size_t) len);
-				secret[MAX_SECRET_LENGTH] = '\0';
-				return 0;
 			}
 		}
 	}
@@ -816,7 +795,7 @@ int rc_find_server (rc_handle const *rh, char const *server_name, uint32_t *ip_a
 	if ((clientfd = fopen (rc_conf_str(rh, "servers"), "r")) == NULL)
 	{
 		rc_log(LOG_ERR, "rc_find_server: couldn't open file: %s: %s", strerror(errno), rc_conf_str(rh, "servers"));
-		return -1;
+		goto fail;
 	}
 
 	while (fgets (buffer, sizeof (buffer), clientfd) != NULL)
@@ -827,53 +806,53 @@ int rc_find_server (rc_handle const *rh, char const *server_name, uint32_t *ip_a
 		if ((h = strtok_r(buffer, " \t\n", &buffer_save)) == NULL) /* first hostname */
 			continue;
 
-		memset (hostnm, '\0', AUTH_ID_LEN);
-		len = strlen (h);
-		if (len > AUTH_ID_LEN)
-		{
-			len = AUTH_ID_LEN;
-		}
-		strncpy (hostnm, h, (size_t) len);
-		hostnm[AUTH_ID_LEN] = '\0';
+		strlcpy (hostnm, h, AUTH_ID_LEN);
 
 		if ((s = strtok_r (NULL, " \t\n", &buffer_save)) == NULL) /* and secret field */
 			continue;
 
-		memset (secret, '\0', MAX_SECRET_LENGTH);
-		len = strlen (s);
-		if (len > MAX_SECRET_LENGTH)
-		{
-			len = MAX_SECRET_LENGTH;
-		}
-		strncpy (secret, s, (size_t) len);
-		secret[MAX_SECRET_LENGTH] = '\0';
+		strlcpy (secret, s, MAX_SECRET_LENGTH);
 
 		if (!strchr (hostnm, '/')) /* If single name form */
 		{
-			if (find_match (ip_addr, hostnm) == 0)
+			tmpinfo = rc_getaddrinfo(hostnm, 0);
+			if (tmpinfo)
 			{
-				result++;
-				break;
+				result = find_match (*info, tmpinfo);
+				if (result == 0)
+				{
+					result++;
+					break;
+				}
+
+				freeaddrinfo(tmpinfo);
+				tmpinfo = NULL;
 			}
 		}
 		else /* <name1>/<name2> "paired" form */
 		{
 			strtok_r(hostnm, "/", &hostnm_save);
-			if (rc_is_myname(hostnm) == 0)
-			{	     /* If we're the 1st name, target is 2nd */
-				if (find_match (ip_addr, hostnm_save) == 0)
-				{
-					result++;
-					break;
-				}
-			}
-			else	/* If we were 2nd name, target is 1st name */
+			tmpinfo = rc_getaddrinfo(hostnm, 0);
+			if (tmpinfo)
 			{
-				if (find_match (ip_addr, hostnm) == 0)
-				{
-					result++;
-					break;
+				if (rc_is_myname(tmpinfo) == 0)
+				{	     /* If we're the 1st name, target is 2nd */
+					if (find_match (*info, tmpinfo) == 0)
+					{
+						result++;
+						break;
+					}
 				}
+				else	/* If we were 2nd name, target is 1st name */
+				{
+					if (find_match (*info, tmpinfo) == 0)
+					{
+						result++;
+						break;
+					}
+				}
+				freeaddrinfo(tmpinfo);
+				tmpinfo = NULL;
 			}
 		}
 	}
@@ -884,17 +863,29 @@ int rc_find_server (rc_handle const *rh, char const *server_name, uint32_t *ip_a
 		memset (secret, '\0', MAX_SECRET_LENGTH);
 		rc_log(LOG_ERR, "rc_find_server: couldn't find RADIUS server %s in %s",
 			 server_name, rc_conf_str(rh, "servers"));
-		return -1;
+		goto fail;
 	}
-	return 0;
+	
+	result = 0;
+	goto cleanup;
+
+ fail:
+ 	freeaddrinfo(*info);
+ 	result = -1;
+
+ cleanup:
+ 	if (tmpinfo)
+ 		freeaddrinfo(tmpinfo);
+
+	return result;
 }
 
-/*
- * Function: rc_config_free
+/**
+ * rc_config_free:
+ * @param rh a handle to parsed configuration
  *
- * Purpose: Free allocated config values
+ * Free allocated config values
  *
- * Arguments: Radius Client handle
  */
 
 void
