@@ -18,15 +18,15 @@
 #include <includes.h>
 #include <freeradius-client.h>
 
-/*
- * Function: rc_read_dictionary
+/** Initialize the dictionary
  *
- * Purpose: Initialize the dictionary.  Read all ATTRIBUTES into
- *	    the dictionary_attributes list.  Read all VALUES into
- *	    the dictionary_values list.
+ * Read all ATTRIBUTES into the dictionary_attributes list.
+ * Read all VALUES into the dictionary_values list.
  *
+ * @param rh a handle to parsed configuration.
+ * @param filename the name of the dictionary file.
+ * @return 0 on success, -1 on failure.
  */
-
 int rc_read_dictionary (rc_handle *rh, char const *filename)
 {
 	FILE           *dictfd;
@@ -42,13 +42,13 @@ int rc_read_dictionary (rc_handle *rh, char const *filename)
 	DICT_VALUE     *dval;
 	DICT_VENDOR    *dvend;
 	char            buffer[256];
-	int             value;
+	int32_t         value;
 	int             type;
-	unsigned attr_vendorspec = 0;
+	uint32_t attr_vendorspec = 0;
 
 	if ((dictfd = fopen (filename, "r")) == NULL)
 	{
-		rc_log(LOG_ERR, "rc_read_dictionary: couldn't open dictionary %s: %s",
+		rc_log(LOG_ERR, "rc_read_dictionary couldn't open dictionary %s: %s",
 				filename, strerror(errno));
 		return -1;
 	}
@@ -104,7 +104,20 @@ int rc_read_dictionary (rc_handle *rh, char const *filename)
 				fclose(dictfd);
 				return -1;
 			}
-			value = atoi (valstr);
+
+			value = atol (valstr);
+
+			/*
+			 *	Disallow values larger than 16 bits, for compatibility.
+			 */
+			if (value > (1L << 16))
+			{
+				rc_log(LOG_ERR,
+                                "rc_read_dictionary: too large value on line %d of dictionary %s",
+                                        line_no, filename);
+                               fclose(dictfd);
+                               return -1;
+			}
 
 			if (strcmp (typestr, "string") == 0)
 			{
@@ -169,14 +182,14 @@ int rc_read_dictionary (rc_handle *rh, char const *filename)
 				return -1;
 			}
 			strcpy (attr->name, namestr);
-			attr->value = value | (attr_vendorspec << 16);
 			attr->type = type;
 
 			if (dvend != NULL) {
-				attr->value = value | (dvend->vendorpec << 16);
+				attr->vendor = dvend->vendorpec;
 			} else {
-				attr->value = value | (attr_vendorspec << 16);
+				attr->vendor = attr_vendorspec;
 			}
+			attr->value = value;
 
 			/* Insert it into the list */
 			attr->next = rh->dictionary_attributes;
@@ -224,7 +237,7 @@ int rc_read_dictionary (rc_handle *rh, char const *filename)
 				fclose(dictfd);
 				return -1;
 			}
-			value = atoi (valstr);
+			value = atol (valstr);
 
 			/* Create a new VALUE entry for the list */
 			if ((dval = malloc (sizeof (DICT_VALUE))) == NULL)
@@ -327,7 +340,7 @@ int rc_read_dictionary (rc_handle *rh, char const *filename)
 				fclose(dictfd);
 				return -1;
 			}
-			value = atoi (valstr);
+			value = atol (valstr);
 
 			/* Create a new VENDOR entry for the list */
 			dvend = malloc(sizeof(DICT_VENDOR));
@@ -349,22 +362,30 @@ int rc_read_dictionary (rc_handle *rh, char const *filename)
 	return 0;
 }
 
-/*
- * Function: rc_dict_getattr
- *
- * Purpose: Return the full attribute structure based on the
- *	    attribute id number.
+/** Lookup a DICT_ATTR by attribute number
  *
+ * @param rh a handle to parsed configuration.
+ * @param attribute the attribute ID.
+ * @return the full attribute structure based on the attribute id number.
  */
-
-DICT_ATTR *rc_dict_getattr (rc_handle const *rh, int attribute)
+DICT_ATTR *rc_dict_getattr(rc_handle const *rh, uint32_t attribute)
 {
 	DICT_ATTR      *attr;
+	uint32_t	vendor = 0;
+
+	/*
+	 *	Allow old-style lookups for compatibility
+	 */
+	if (attribute > (1l << 16)) {
+		vendor = attribute >> 16;
+		attribute &= 0xffff;
+	}
+
 
 	attr = rh->dictionary_attributes;
 	while (attr != NULL)
 	{
-		if (attr->value == attribute)
+		if (attr->vendor == vendor && attr->value == attribute)
 		{
 			return attr;
 		}
@@ -373,15 +394,36 @@ DICT_ATTR *rc_dict_getattr (rc_handle const *rh, int attribute)
 	return NULL;
 }
 
-/*
- * Function: rc_dict_findattr
- *
- * Purpose: Return the full attribute structure based on the
- *	    attribute name.
+/** Lookup a DICT_ATTR by attribute number
  *
+ * @param rh a handle to parsed configuration.
+ * @param attribute the attribute ID.
+ * @param vendor the vendor ID.
+ * @return the full attribute structure based on the attribute id number.
  */
+DICT_ATTR *rc_dict_get_vendor_attr(rc_handle const *rh, uint32_t attribute, uint32_t vendor)
+{
+	DICT_ATTR      *attr;
 
-DICT_ATTR *rc_dict_findattr (rc_handle const *rh, char const *attrname)
+	attr = rh->dictionary_attributes;
+	while (attr != NULL)
+	{
+		if (attr->vendor == vendor && attr->value == attribute)
+		{
+			return attr;
+		}
+		attr = attr->next;
+	}
+	return NULL;
+}
+/** Lookup a DICT_ATTR by its name
+ *
+ * @param rh a handle to parsed configuration.
+ * @param attrname the attribute name.
+ *
+ * @return the full attribute structure based on the attribute name.
+ */
+DICT_ATTR *rc_dict_findattr(rc_handle const *rh, char const *attrname)
 {
 	DICT_ATTR      *attr;
 
@@ -398,15 +440,13 @@ DICT_ATTR *rc_dict_findattr (rc_handle const *rh, char const *attrname)
 }
 
 
-/*
- * Function: rc_dict_findval
- *
- * Purpose: Return the full value structure based on the
- *         value name.
+/** Lookup a DICT_VALUE by its name
  *
+ * @param rh a handle to parsed configuration.
+ * @param valname the value name.
+ * @return the full value structure based on the value name.
  */
-
-DICT_VALUE *rc_dict_findval (rc_handle const *rh, char const *valname)
+DICT_VALUE *rc_dict_findval(rc_handle const *rh, char const *valname)
 {
 	DICT_VALUE     *val;
 
@@ -422,16 +462,13 @@ DICT_VALUE *rc_dict_findval (rc_handle const *rh, char const *valname)
 	return NULL;
 }
 
-/*
- * Function: rc_dict_findvend
- *
- * Purpose: Return the full vendor structure based on the
- *          vendor name.
+/** Lookup a DICT_VENDOR by its name
  *
+ * @param rh a handle to parsed configuration.
+ * @param valname the vendor name.
+ * @return the full vendor structure based on the vendor name.
  */
-
-DICT_VENDOR *
-rc_dict_findvend(rc_handle const *rh, char const *vendorname)
+DICT_VENDOR *rc_dict_findvend(rc_handle const *rh, char const *vendorname)
 {
 	DICT_VENDOR	*vend;
 
@@ -441,16 +478,13 @@ rc_dict_findvend(rc_handle const *rh, char const *vendorname)
 	return NULL;
 }
 
-/*
- * Function: rc_dict_getvend
- *
- * Purpose: Return the full vendor structure based on the
- *          vendor id number.
+/** Lookup a DICT_VENDOR by its IANA number
  *
+ * @param rh a handle to parsed configuration.
+ * @param vendorpec the vendor ID.
+ * @return the full vendor structure based on the vendor id number.
  */
-
-DICT_VENDOR *
-rc_dict_getvend (rc_handle const *rh, int vendorpec)
+DICT_VENDOR *rc_dict_getvend (rc_handle const *rh, uint32_t vendorpec)
 {
         DICT_VENDOR      *vend;
 
@@ -460,16 +494,14 @@ rc_dict_getvend (rc_handle const *rh, int vendorpec)
 	return NULL;
 }
 
-/*
- * Function: dict_getval
- *
- * Purpose: Return the full value structure based on the
- *          actual value and the associated attribute name.
+/** Get DICT_VALUE based on attribute name and integer value number
  *
+ * @param rh a handle to parsed configuration.
+ * @param value the attribute value.
+ * @param attrname the attribute name.
+ * @return the full value structure based on the actual value and the associated attribute name.
  */
-
-DICT_VALUE *
-rc_dict_getval (rc_handle const *rh, uint32_t value, char const *attrname)
+DICT_VALUE *rc_dict_getval(rc_handle const *rh, uint32_t value, char const *attrname)
 {
 	DICT_VALUE     *val;
 
@@ -486,16 +518,11 @@ rc_dict_getval (rc_handle const *rh, uint32_t value, char const *attrname)
 	return NULL;
 }
 
-/*
- * Function: rc_dict_free
+/** Frees the allocated av lists
  *
- * Purpose: Free allocated av lists
- *
- * Arguments: Radius Client handle
+ * @param rh a handle to parsed configuration.
  */
-
-void
-rc_dict_free(rc_handle *rh)
+void rc_dict_free(rc_handle *rh)
 {
 	DICT_ATTR	*attr, *nattr;
 	DICT_VALUE	*val, *nval;
