@@ -21,20 +21,24 @@
 #include <freeradius-client.h>
 #include <pathnames.h>
 
+#if (defined(__APPLE__) || defined(__FreeBSD__)) && defined(HAVE_GETENTROPY)
+# include <sys/random.h>
+#endif
+
+#include "util.h"
+
 #define	SA(p)	((struct sockaddr *)(p))
 
 static void rc_random_vector (unsigned char *);
 static int rc_check_reply (AUTH_HDR *, int, char const *, unsigned char const *, unsigned char);
 
-/*
- * Function: rc_pack_list
- *
- * Purpose: Packs an attribute value pair list into a buffer.
- *
- * Returns: Number of octets packed.
+/** Packs an attribute value pair list into a buffer
  *
+ * @param vp a pointer to a #VALUE_PAIR.
+ * @param secret the secret used by the server.
+ * @param auth a pointer to #AUTH_HDR.
+ * @return The number of octets packed.
  */
-
 static int rc_pack_list (VALUE_PAIR *vp, char *secret, AUTH_HDR *auth)
 {
 	int             length, i, pc, padded_length;
@@ -50,15 +54,18 @@ static int rc_pack_list (VALUE_PAIR *vp, char *secret, AUTH_HDR *auth)
 	while (vp != NULL)
 	{
 		vsa_length_ptr = NULL;
-		if (VENDOR(vp->attribute) != 0) {
+		if (vp->vendor != 0) {
 			*buf++ = PW_VENDOR_SPECIFIC;
 			vsa_length_ptr = buf;
 			*buf++ = 6;
-			vendor = htonl(VENDOR(vp->attribute));
+			vendor = htonl(vp->vendor);
 			memcpy(buf, &vendor, sizeof(uint32_t));
 			buf += 4;
 			total_length += 6;
 		}
+
+		if (vp->attribute > 0xff) continue;
+
 		*buf++ = (vp->attribute & 0xff);
 
 		switch (vp->attribute)
@@ -106,37 +113,6 @@ static int rc_pack_list (VALUE_PAIR *vp, char *secret, AUTH_HDR *auth)
 		  total_length += padded_length + 2;
 
 		  break;
-#if 0
-		 case PW_CHAP_PASSWORD:
-
-		  *buf++ = CHAP_VALUE_LENGTH + 2;
-		  if (vsa_length_ptr != NULL) *vsa_length_ptr += CHAP_VALUE_LENGTH + 2;
-
-		  /* Encrypt the Password */
-		  length = vp->lvalue;
-		  if (length > CHAP_VALUE_LENGTH)
-		  {
-			length = CHAP_VALUE_LENGTH;
-		  }
-		  memset ((char *) passbuf, '\0', CHAP_VALUE_LENGTH);
-		  memcpy ((char *) passbuf, vp->strvalue, (size_t) length);
-
-		  /* Calculate the MD5 Digest */
-		  secretlen = strlen (secret);
-		  strcpy ((char *) md5buf, secret);
-		  memcpy ((char *) md5buf + secretlen, (char *) auth->vector,
-		  	  AUTH_VECTOR_LEN);
-		  rc_md5_calc (buf, md5buf, secretlen + AUTH_VECTOR_LEN);
-
-		  /* Xor the password into the MD5 digest */
-		  for (i = 0; i < CHAP_VALUE_LENGTH; i++)
-		  {
-			*buf++ ^= passbuf[i];
-		  }
-		  total_length += CHAP_VALUE_LENGTH + 2;
-
-		  break;
-#endif
 		 default:
 		  switch (vp->type)
 		  {
@@ -151,6 +127,7 @@ static int rc_pack_list (VALUE_PAIR *vp, char *secret, AUTH_HDR *auth)
 
 		    case PW_TYPE_IPV6ADDR:
 			length = 16;
+			*buf++ = length + 2;
 			if (vsa_length_ptr != NULL) *vsa_length_ptr += length + 2;
 			memcpy (buf, vp->strvalue, (size_t) length);
 			buf += length;
@@ -159,6 +136,7 @@ static int rc_pack_list (VALUE_PAIR *vp, char *secret, AUTH_HDR *auth)
 
 		    case PW_TYPE_IPV6PREFIX:
 			length = vp->lvalue;
+			*buf++ = length + 2;
 			if (vsa_length_ptr != NULL) *vsa_length_ptr += length + 2;
 			memcpy (buf, vp->strvalue, (size_t) length);
 			buf += length;
@@ -186,9 +164,12 @@ static int rc_pack_list (VALUE_PAIR *vp, char *secret, AUTH_HDR *auth)
 	return total_length;
 }
 
-/* Function strappend
+/** Appends a string to the provided buffer
  *
- * Purpose: appends a string to the provided buffer
+ * @param dest the destination buffer.
+ * @param max_size the maximum size available in the destination buffer.
+ * @param pos the current position in the dest buffer; initially must be zero.
+ * @param src the source buffer to append.
  */
 static void strappend(char *dest, unsigned max_size, int *pos, const char *src)
 {
@@ -207,31 +188,37 @@ static void strappend(char *dest, unsigned max_size, int *pos, const char *src)
 	return;
 }
 
-/*
- * Function: rc_send_server
- *
- * Purpose: send a request to a RADIUS server and wait for the reply
+/** Sends a request to a RADIUS server and waits for the reply
  *
+ * @param rh a handle to parsed configuration
+ * @param data a pointer to a #SEND_DATA structure
+ * @param msg must be an array of %PW_MAX_MSG_SIZE or %NULL; will contain the concatenation of
+ *	any %PW_REPLY_MESSAGE received.
+ * @param flags must be %AUTH or %ACCT
+ * @return %OK_RC (0) on success, %TIMEOUT_RC on timeout %REJECT_RC on acess reject, or negative
+ *	on failure as return value.
  */
-
-int rc_send_server (rc_handle *rh, SEND_DATA *data, char *msg)
+int rc_send_server (rc_handle const *rh, SEND_DATA *data, char *msg, unsigned flags)
 {
 	int             sockfd;
-	struct sockaddr_in sinlocal;
-	struct sockaddr_in sinremote;
 	AUTH_HDR       *auth, *recv_auth;
-	uint32_t           auth_ipaddr, nas_ipaddr;
 	char           *server_name;	/* Name of server to query */
+	struct sockaddr_storage our_sockaddr;
+	struct addrinfo *auth_addr = NULL;
 	socklen_t       salen;
 	int             result = 0;
 	int             total_length;
 	int             length, pos;
 	int             retry_max;
-	size_t			secretlen;
+	unsigned	discover_local_ip;
+	size_t		secretlen;
 	char            secret[MAX_SECRET_LENGTH + 1];
 	unsigned char   vector[AUTH_VECTOR_LEN];
-	char            recv_buffer[BUFFER_LEN];
-	char            send_buffer[BUFFER_LEN];
+	uint8_t          recv_buffer[BUFFER_LEN];
+	uint8_t          send_buffer[BUFFER_LEN];
+	char		our_addr_txt[50]; /* hold a text IP */
+	char		auth_addr_txt[50]; /* hold a text IP */
+	uint8_t		*attr;
 	int		retries;
 	VALUE_PAIR 	*vp;
 	struct pollfd	pfd;
@@ -241,75 +228,94 @@ int rc_send_server (rc_handle *rh, SEND_DATA *data, char *msg)
 	if (server_name == NULL || server_name[0] == '\0')
 		return ERROR_RC;
 
-	if ((vp = rc_avpair_get(data->send_pairs, PW_SERVICE_TYPE, 0)) && \
-	    (vp->lvalue == PW_ADMINISTRATIVE))
+	if(data->secret != NULL)
 	{
-		strcpy(secret, MGMT_POLL_SECRET);
-		if ((auth_ipaddr = rc_get_ipaddr(server_name)) == 0)
+		// no need to look up the secret from configuration
+		strlcpy(secret, data->secret, MAX_SECRET_LENGTH);
+		auth_addr = rc_getaddrinfo (server_name, flags==AUTH?PW_AI_AUTH:PW_AI_ACCT);
+		if(auth_addr == NULL)
+		{
+			rc_log(LOG_ERR, "rc_send_server: unable to resolve server: %s", server_name);
 			return ERROR_RC;
+		}
 	}
-	else
+	else if (rc_find_server_addr (rh, server_name, &auth_addr, secret, flags) != 0)
 	{
-		if(data->secret != NULL)
-		{
-			strncpy(secret, data->secret, MAX_SECRET_LENGTH);
+		rc_log(LOG_ERR, "rc_send_server: unable to find server: %s", server_name);
+		return ERROR_RC;
+	}
+
+	rc_own_bind_addr(rh, &our_sockaddr);
+	discover_local_ip = 0;
+	if (our_sockaddr.ss_family == AF_INET) {
+		if (((struct sockaddr_in*)(&our_sockaddr))->sin_addr.s_addr == INADDR_ANY) {
+			discover_local_ip = 1;
 		}
-		/*
-		else
-		{
-		*/
-		if (rc_find_server (rh, server_name, &auth_ipaddr, secret) != 0)
-		{
-			rc_log(LOG_ERR, "rc_send_server: unable to find server: %s", server_name);
-			return ERROR_RC;
-		}
-		/*}*/
 	}
 
 	DEBUG(LOG_ERR, "DEBUG: rc_send_server: creating socket to: %s", server_name);
+	if (discover_local_ip) {
+		result = rc_get_srcaddr(SA(&our_sockaddr), auth_addr->ai_addr);
+		if (result != 0) {
+			memset (secret, '\0', sizeof (secret));
+			rc_log(LOG_ERR, "rc_send_server: cannot figure our own address");
+			result = ERROR_RC;
+			goto cleanup;
+		}
+	}
 
-	sockfd = socket (AF_INET, SOCK_DGRAM, 0);
+	sockfd = socket (our_sockaddr.ss_family, SOCK_DGRAM, 0);
 	if (sockfd < 0)
 	{
 		memset (secret, '\0', sizeof (secret));
 		rc_log(LOG_ERR, "rc_send_server: socket: %s", strerror(errno));
-		return ERROR_RC;
+		result = ERROR_RC;
+		goto cleanup;
 	}
 
-	memset((char *)&sinlocal, '\0', sizeof(sinlocal));
-	sinlocal.sin_family = AF_INET;
-	sinlocal.sin_addr.s_addr = htonl(rc_own_bind_ipaddress(rh));
-	sinlocal.sin_port = htons((unsigned short) 0);
-	if (bind(sockfd, SA(&sinlocal), sizeof(sinlocal)) < 0)
+	if (our_sockaddr.ss_family == AF_INET)
+		((struct sockaddr_in*)&our_sockaddr)->sin_port = 0;
+	else
+		((struct sockaddr_in6*)&our_sockaddr)->sin6_port = 0;
+
+	if (bind(sockfd, SA(&our_sockaddr), SS_LEN(&our_sockaddr)) < 0)
 	{
 		close (sockfd);
 		memset (secret, '\0', sizeof (secret));
 		rc_log(LOG_ERR, "rc_send_server: bind: %s: %s", server_name, strerror(errno));
-		return ERROR_RC;
+		result = ERROR_RC;
+		goto cleanup;
 	}
 
 	retry_max = data->retries;	/* Max. numbers to try for reply */
 	retries = 0;			/* Init retry cnt for blocking call */
 
-	memset ((char *)&sinremote, '\0', sizeof(sinremote));
-	sinremote.sin_family = AF_INET;
-	sinremote.sin_addr.s_addr = htonl (auth_ipaddr);
-	sinremote.sin_port = htons ((unsigned short) data->svc_port);
+	if (data->svc_port) {
+		if (our_sockaddr.ss_family == AF_INET)
+			((struct sockaddr_in*)auth_addr->ai_addr)->sin_port = htons ((unsigned short) data->svc_port);
+		else
+			((struct sockaddr_in6*)auth_addr->ai_addr)->sin6_port = htons ((unsigned short) data->svc_port);
+	}
 
 	/*
 	 * Fill in NAS-IP-Address (if needed)
 	 */
-	if (rc_avpair_get(data->send_pairs, PW_NAS_IP_ADDRESS, 0) == NULL) {
-		if (sinlocal.sin_addr.s_addr == htonl(INADDR_ANY)) {
-			if (rc_get_srcaddr(SA(&sinlocal), SA(&sinremote)) != 0) {
-				close (sockfd);
-				memset (secret, '\0', sizeof (secret));
-				return ERROR_RC;
-			}
+	if (rc_avpair_get(data->send_pairs, PW_NAS_IP_ADDRESS, 0) == NULL &&
+	    rc_avpair_get(data->send_pairs, PW_NAS_IPV6_ADDRESS, 0) == NULL) {
+		if (our_sockaddr.ss_family == AF_INET) {
+			uint32_t ip;
+			ip = *((uint32_t*)(&((struct sockaddr_in*)&our_sockaddr)->sin_addr));
+			ip = ntohl(ip);
+
+			rc_avpair_add(rh, &(data->send_pairs), PW_NAS_IP_ADDRESS,
+			    &ip, 0, 0);
+		} else {
+			void *p;
+			p = &((struct sockaddr_in6*)&our_sockaddr)->sin6_addr;
+
+			rc_avpair_add(rh, &(data->send_pairs), PW_NAS_IPV6_ADDRESS,
+			    p, 0, 0);
 		}
-		nas_ipaddr = ntohl(sinlocal.sin_addr.s_addr);
-		rc_avpair_add(rh, &(data->send_pairs), PW_NAS_IP_ADDRESS,
-		    &nas_ipaddr, 0, 0);
 	}
 
 	/* Build a request */
@@ -339,21 +345,28 @@ int rc_send_server (rc_handle *rh, SEND_DATA *data, char *msg)
 		auth->length = htons ((unsigned short) total_length);
 	}
 
+	getnameinfo(SA(&our_sockaddr), SS_LEN(&our_sockaddr), NULL, 0, our_addr_txt, sizeof(our_addr_txt), NI_NUMERICHOST);
+	getnameinfo(auth_addr->ai_addr, auth_addr->ai_addrlen, NULL, 0, auth_addr_txt, sizeof(auth_addr_txt), NI_NUMERICHOST);
+
 	DEBUG(LOG_ERR, "DEBUG: local %s : 0, remote %s : %u\n", 
-		inet_ntoa(sinlocal.sin_addr),
-		inet_ntoa(sinremote.sin_addr), data->svc_port);
+	      our_addr_txt, auth_addr_txt, data->svc_port);
 
 	for (;;)
 	{
-		sendto (sockfd, (char *) auth, (unsigned int) total_length, (int) 0,
-			SA(&sinremote), sizeof (struct sockaddr_in));
+		do {
+			result = sendto (sockfd, (char *) auth, (unsigned int)total_length, 
+				(int) 0, SA(auth_addr->ai_addr), auth_addr->ai_addrlen);
+		} while (result == -1 && errno == EINTR);
+		if (result == -1) {
+			rc_log(LOG_ERR, "%s: socket: %s", __FUNCTION__, strerror(errno));
+		}
 
 		pfd.fd = sockfd;
 		pfd.events = POLLIN;
 		pfd.revents = 0;
-		start_time = rc_getctime();
+		start_time = rc_getmtime();
 		for (timeout = data->timeout; timeout > 0;
-		    timeout -= rc_getctime() - start_time) {
+		    timeout -= rc_getmtime() - start_time) {
 			result = poll(&pfd, 1, timeout * 1000);
 			if (result != -1 || errno != EINTR)
 				break;
@@ -363,7 +376,8 @@ int rc_send_server (rc_handle *rh, SEND_DATA *data, char *msg)
 			rc_log(LOG_ERR, "rc_send_server: poll: %s", strerror(errno));
 			memset (secret, '\0', sizeof (secret));
 			close (sockfd);
-			return ERROR_RC;
+			result = ERROR_RC;
+			goto cleanup;
 		}
 		if (result == 1 && (pfd.revents & POLLIN) != 0)
 			break;
@@ -375,17 +389,20 @@ int rc_send_server (rc_handle *rh, SEND_DATA *data, char *msg)
 		if (retries++ >= retry_max)
 		{
 			rc_log(LOG_ERR,
-				"rc_send_server: no reply from RADIUS server %s:%u, %s",
-				 rc_ip_hostname (auth_ipaddr), data->svc_port, inet_ntoa(sinremote.sin_addr));
+				"rc_send_server: no reply from RADIUS server %s:%u",
+				 auth_addr_txt, data->svc_port);
 			close (sockfd);
 			memset (secret, '\0', sizeof (secret));
-			return TIMEOUT_RC;
+			result = TIMEOUT_RC;
+			goto cleanup;
 		}
 	}
-	salen = sizeof(sinremote);
-	length = recvfrom (sockfd, (char *) recv_buffer,
-			   (int) sizeof (recv_buffer),
-			   (int) 0, SA(&sinremote), &salen);
+	salen = auth_addr->ai_addrlen;
+	do {
+		length = recvfrom (sockfd, (char *) recv_buffer,
+				   (int) sizeof (recv_buffer),
+				   (int) 0, SA(auth_addr->ai_addr), &salen);
+	} while(length == -1 && errno == EINTR);
 
 	if (length <= 0)
 	{
@@ -393,7 +410,8 @@ int rc_send_server (rc_handle *rh, SEND_DATA *data, char *msg)
 			 data->svc_port, strerror(errno));
 		close (sockfd);
 		memset (secret, '\0', sizeof (secret));
-		return ERROR_RC;
+		result = ERROR_RC;
+		goto cleanup;
 	}
 
 	recv_auth = (AUTH_HDR *)recv_buffer;
@@ -403,7 +421,45 @@ int rc_send_server (rc_handle *rh, SEND_DATA *data, char *msg)
 		    server_name, data->svc_port);
 		close(sockfd);
 		memset(secret, '\0', sizeof(secret));
-		return ERROR_RC;
+		result = ERROR_RC;
+		goto cleanup;
+	}
+
+	/*
+	 *	If UDP is larger than RADIUS, shorten it to RADIUS.
+	 */
+	if (length > ntohs(recv_auth->length)) length = ntohs(recv_auth->length);
+
+	/*
+	 *	Verify that it's a valid RADIUS packet before doing ANYTHING with it.
+	 */
+	attr = recv_buffer + AUTH_HDR_LEN;
+	while (attr < (recv_buffer + length)) {
+		if (attr[0] == 0) {
+			rc_log(LOG_ERR, "rc_send_server: recvfrom: %s:%d: attribute zero is invalid",
+			       server_name, data->svc_port);
+			close(sockfd);
+			memset(secret, '\0', sizeof(secret));
+			return ERROR_RC;
+		}
+
+		if (attr[1] < 2) {
+			rc_log(LOG_ERR, "rc_send_server: recvfrom: %s:%d: attribute length is too small",
+			       server_name, data->svc_port);
+			close(sockfd);
+			memset(secret, '\0', sizeof(secret));
+			return ERROR_RC;
+		}
+
+		if ((attr + attr[1]) > (recv_buffer + length)) {
+			rc_log(LOG_ERR, "rc_send_server: recvfrom: %s:%d: attribute overflows the packet",
+			       server_name, data->svc_port);
+			close(sockfd);
+			memset(secret, '\0', sizeof(secret));
+			return ERROR_RC;
+		}
+
+		attr += attr[1];
 	}
 
 	result = rc_check_reply (recv_auth, BUFFER_LEN, secret, vector, data->seq_nbr);
@@ -419,7 +475,9 @@ int rc_send_server (rc_handle *rh, SEND_DATA *data, char *msg)
 	close (sockfd);
 	memset (secret, '\0', sizeof (secret));
 
-	if (result != OK_RC) return result;
+	if (result != OK_RC) {
+		goto cleanup;
+	}
 
 	if (msg) {
 		*msg = '\0';
@@ -449,22 +507,26 @@ int rc_send_server (rc_handle *rh, SEND_DATA *data, char *msg)
 	}
 	else
 	{
+		rc_log(LOG_ERR, "rc_send_server: received RADIUS server response neither ACCEPT nor REJECT, invalid");
 		result = BADRESP_RC;
 	}
 
+ cleanup:
+ 	if (auth_addr)
+ 		freeaddrinfo(auth_addr);
+
 	return result;
 }
 
-/*
- * Function: rc_check_reply
- *
- * Purpose: verify items in returned packet.
- *
- * Returns:	OK_RC       -- upon success,
- *		BADRESP_RC  -- if anything looks funny.
+/** Verify items in returned packet
  *
+ * @param auth a pointer to #AUTH_HDR.
+ * @param bufferlen the available buffer length.
+ * @param secret the secret used by the server.
+ * @param vector a random vector of %AUTH_VECTOR_LEN.
+ * @param seq_nbr a unique sequence number.
+ * @return %OK_RC upon success, %BADRESP_RC if anything looks funny.
  */
-
 static int rc_check_reply (AUTH_HDR *auth, int bufferlen, char const *secret, unsigned char const *vector, uint8_t seq_nbr)
 {
 	int             secretlen;
@@ -551,18 +613,6 @@ static int rc_check_reply (AUTH_HDR *auth, int bufferlen, char const *secret, un
 	if (memcmp ((char *) reply_digest, (char *) calc_digest,
 		    AUTH_VECTOR_LEN) != 0)
 	{
-#ifdef RADIUS_116
-		/* the original Livingston radiusd v1.16 seems to have
-		   a bug in digest calculation with accounting requests,
-		   authentication request are ok. i looked at the code
-		   but couldn't find any bugs. any help to get this
-		   kludge out are welcome. preferably i want to
-		   reproduce the calculation bug here to be compatible
-		   to stock Livingston radiusd v1.16.	-lf, 03/14/96
-		 */
-		if (auth->code == PW_ACCOUNTING_RESPONSE)
-			return OK_RC;
-#endif
 		rc_log(LOG_ERR, "rc_check_reply: received invalid reply digest from RADIUS server");
 		return BADRESP_RC;
 	}
@@ -571,15 +621,10 @@ static int rc_check_reply (AUTH_HDR *auth, int bufferlen, char const *secret, un
 
 }
 
-/*
- * Function: rc_random_vector
- *
- * Purpose: generates a random vector of AUTH_VECTOR_LEN octets.
- *
- * Returns: the vector (call by reference)
+/** Generates a random vector of AUTH_VECTOR_LEN octets
  *
+ * @param vector a buffer with at least %AUTH_VECTOR_LEN bytes.
  */
-
 static void rc_random_vector (unsigned char *vector)
 {
 	int             randno;
@@ -621,7 +666,7 @@ static void rc_random_vector (unsigned char *vector)
  fallback:
 	for (i = 0; i < AUTH_VECTOR_LEN;)
 	{
-		randno = random ();
+		randno = rc_random();
 		memcpy ((char *) vector, (char *) &randno, sizeof (int));
 		vector += sizeof (int);
 		i += sizeof (int);
